#include "config.h"

#include "ds485-socket-server.h"

#include <syslog.h>
#include <pthread.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/uio.h>

#include <utlist.h>

#include <digitalSTROM/ds485.h>
#include <digitalSTROM/ds485-core.h>
#include "ds485-socket-interface.h"

#include "ds485-socket-server-impl.h"
#include "base/log.h"
#include "base/util.h"
#include "base/socket_send.h"

/* CAUTION: client->lock must be held */
int _ds485n_tlv_write(dS485dClient *cl, uint8_t cmd, const void *data, int dlen)
{
    uint16_t length = dlen + 3;
    struct iovec iov[2];
    uint8_t buf[8];
    int rval;

    buf[0] = cmd;
    buf[1] = length & 0xff;
    buf[2] = (length >> 8) & 0xff;

    iov[0].iov_base = buf;
    iov[0].iov_len = 3;
    iov[1].iov_base = (void *)data;
    iov[1].iov_len = dlen;

    if (cl->fd < 0) {
        return -EBADF;
    }
    rval = socket_writev(cl->fd, iov, 2);

    if (rval) {
        report2(LOG_NOTICE, cl->peer_name, "_ds485n_tlv_write: %d %s\n",
                cmd, strerror(rval));
        _ds485n_client_detach(cl);
        return rval;
    }

    return 0;
}

/* CAUTION: client->lock must be held */
int _ds485n_dsmapi_write(dS485dClient* cl, uint8_t cmd,
                         const ds485_container_t *container)
{
    /* TODO see dS485CmdApidata, maybe remove it */
    return _ds485n_tlv_write(cl, cmd, container,
                             DS485_CONTAINER_SIZE + container->length);
}

void ds485_server_emit_bus_change(struct dS485_server *server, dsuid_t dsuid, int flags)
{
    dS485dClient* cs;
    ds485_container_t container;
    int rval;

    ds485n_create_bus_change(&container, &dsuid, flags);

    pthread_rwlock_rdlock(&server->clientListRWLock);
    LL_FOREACH(server->clSessions, cs) {

        pthread_mutex_lock(&cs->lock);
        if (cs->fd < 0) {
            /* ignore leaving clients */
            pthread_mutex_unlock(&cs->lock);
            continue;
        }

        rval = _ds485n_dsmapi_write(cs, DS485D_CMD_BUS_CHANGE, &container);
        pthread_mutex_unlock(&cs->lock);

        if (rval) {
            report2(LOG_NOTICE, cs->peer_name, "missed bus change err:%s\n",
                    strerror(-rval));
            continue;
        }
    }
    pthread_rwlock_unlock(&server->clientListRWLock);
}
NETLIB_EXPORTED(ds485_server_emit_bus_change);

void ds485_server_emit_bus_state(struct dS485_server *server, bus_state_t state)
{
    dS485dClient* cs;
    ds485_container_t container;
    int rval;

    ds485n_create_bus_state(&container, state);

    pthread_rwlock_rdlock(&server->clientListRWLock);
    LL_FOREACH(server->clSessions, cs) {

        pthread_mutex_lock(&cs->lock);
        if (cs->fd < 0) {
            /* ignore leaving clients */
            pthread_mutex_unlock(&cs->lock);
            continue;
        }

        rval = _ds485n_dsmapi_write(cs, DS485D_CMD_BUS_STATE, &container);
        pthread_mutex_unlock(&cs->lock);

        if (rval) {
            report2(LOG_NOTICE, cs->peer_name, "missed state change err:%s\n",
                    strerror(-rval));
            continue;
        }
    }
    pthread_rwlock_unlock(&server->clientListRWLock);
}
NETLIB_EXPORTED(ds485_server_emit_bus_state);

/**
 * Sends reply to client denoted by transaction id
 *
 * Problems of the routing protocol ds485-netlib v1.
 * - clients do not have an address
 * - netlib assigns clients a transactionId upon connect
 *   and sets source address to be the address of the server
 *
 * The transactionId is abused to tag the container with the
 * client connection. A reply is only sent to the client with the
 * matching transactionId.
 * In a ds485p / ds485 stacked setup. 1st the ds485p changes the
 * transactionId of the container, then ds485d again. So the
 * ds485p transactionId is lost, and ds485p can't rely on it.
 *
 * The solution to this was to split the transaction id into
 * halves, ds485p is using th upper 4 bits, vdsm/ds485d the lower
 * 4 bits. This is the reason why you can't connect a ds485p to
 * another ds485p, they would both overwrite the same
 * transactionId and replies would be lost
 *
 * packet is forwarded
 * - to client with matching transactionId
 * - all sniffing clients
 * - all clients if message is event or request
 */
static void ds485n_pu_reply(struct dS485_server *server,
                                const packet_union_t *packet)
{
    dS485dClient* cs;
    PKT_UNION_STR_BUF(buf);
    bool delivered = false;
    int forwardId;
    int rval;

    if (ds485n_pu_is_oversize(packet)) {
        sprintf_packet_union(buf, packet);
        report(LOG_WARNING, "reply: oversize packet: %s\n", buf);
        return;
    }

    forwardId = (server->upperPath)
        ? packet->transactionId >> 4
        : packet->transactionId & 0x0f;

    server->rxPacketNum++;

    pthread_rwlock_rdlock(&server->clientListRWLock);
    LL_FOREACH(server->clSessions, cs) {

        if ((cs->transactionId == forwardId) ||
            (packet->transactionId == 0) ||
            (packet->containerType != DS485_CONTAINER_RESPONSE) ||
            (cs->rxfilter & DS485_FILTER_ALL_RESPONSES)) {

            pthread_mutex_lock(&cs->lock);
            if (cs->fd < 0) {
                /* ignore leaving clients */
                pthread_mutex_unlock(&cs->lock);
                continue;
            }

            rval = _ds485n_pu_write(cs, packet);
            pthread_mutex_unlock(&cs->lock);

            if (rval) {
                sprintf_packet_union(buf, packet);
                report2(LOG_NOTICE, cs->peer_name, "REPLY err:%s pkt:%s\n",
                        strerror(-rval), buf);
                continue;
            }

            server->txPacketNum++;
            delivered |= (cs->transactionId == forwardId);
        }
    }
    pthread_rwlock_unlock(&server->clientListRWLock);

    if (!delivered && !IsBroadcastDsuid(packet->dst)) {
        sprintf_packet_union(buf, packet);
        report(LOG_NOTICE, "REPLY undelivered: %s\n", buf);
    }
}

void ds485_server_reply(struct dS485_server *server,
                        const ds485_container_t *container)
{
    const packet_union_t data;
    ds485n_pu_from_short((packet_union_t *)&data, container);
    return ds485n_pu_reply(server, &data);
}
NETLIB_EXPORTED(ds485_server_reply);

void ds485_server_reply_long(struct dS485_server *server,
                             const ds485n_packet_t *lcont)
{
    const packet_union_t data;
    ds485n_pu_from_long((packet_union_t *)&data, lcont);
    return ds485n_pu_reply(server, &data);
}
NETLIB_EXPORTED(ds485_server_reply_long);

/**
 * Forward packet to server and all peers incl. ourselves if promiscous
 */
int ds485n_pu_request(struct dS485_server *server, packet_union_t *packet,
                      dS485dClient *sender)
{
    PKT_UNION_STR_BUF(buf);
    dS485dClient* cs;
    int ret;

    /* netlib_packet_parser always passes sender */
    assert(sender);

    if (ds485n_pu_is_oversize(packet)) {
        sprintf_packet_union(buf, packet);
        report(LOG_WARNING, "request: oversize packet: %s\n", buf);
        return -ENOBUFS;
    }

    if (server->upperPath) {
        packet->transactionId &= 0x0f;
        packet->transactionId |= (sender->transactionId << 4);
    } else {
        packet->transactionId &= 0xf0;
        packet->transactionId |= (sender->transactionId & 0x0f);
    }

    // ds485c has no dsuid
    if (IsNullDsuid(packet->src)) {
        packet->src = server->cb.get_dsuid(server);
    }

    ds485n_pu_sync(packet);

    ret = (packet->type == v_container) ?
        server->cb.receive_data(server, packet->u.container) :
        server->cb.receive_long_data(server, packet->u.packet);
    if (ret) {
        sprintf_packet_union(buf, packet);
        report2(LOG_NOTICE, "REQ: srv err:%s bus:%s pkt:%s\n",
                strerror(-ret), sender->peer_name, buf);
        /* reinvent ICMP return ret;  TODO */
    }

    pthread_rwlock_rdlock(&server->clientListRWLock);
    LL_FOREACH(server->clSessions, cs) {
        int rval;
        /* TODO shouldn't this be logical AND */
        if ((cs != sender) || (cs->rxfilter & DS485_FILTER_ALL_RESPONSES)) {

            pthread_mutex_lock(&cs->lock);
            if (cs->fd < 0) {
                /* ignore leaving clients */
                pthread_mutex_unlock(&cs->lock);
                continue;
            }

            rval = _ds485n_pu_write(cs, packet);
            pthread_mutex_unlock(&cs->lock);

            if (rval) {
                sprintf_packet_union(buf, packet);
                report2(LOG_NOTICE, cs->peer_name, "REQ: cl err:%s bus:%s pkt:%s\n",
                        strerror(-rval), sender->peer_name, buf);
                continue;
            }
            server->txPacketNum++;
        }
    }
    pthread_rwlock_unlock(&server->clientListRWLock);
    return ret;
}

struct dS485_server *ds485_server_init(struct dS485_server_callbacks *cb, bool upperPath)
{
    struct dS485_server *tmp;
    const char *env;
    int id, loglevel;

    init_logger("ds485n", LOG_NOTICE, 0);

    env = getenv(DS485_NETLIB_DEBUG);
    if (env) {
        errno = 0;
        loglevel = strtol(env, NULL, 10);
        if (errno) {
            report(LOG_WARNING, "Invalid DS485_NETLIB_DEBUG=%s, ingoring\n",
                   env);
        } else {
            set_log_level(loglevel);
        }
    }

    env = getenv(DS485_SYSLOG);
    if (env) {
        enable_syslog(true);
    }

    tmp = calloc(1, sizeof(struct dS485_server));
    if (!tmp) {
        return NULL;
    }

    tmp->active = true;
    tmp->sfd = -1;
    tmp->upperPath = upperPath;
    memcpy(&tmp->cb, cb, sizeof(struct dS485_server_callbacks));

    pthread_rwlock_init(&tmp->clientListRWLock, NULL);
    for (id = 0; id < MAX_CLIENTS; id++) {
        tmp->transid_pool[id] = id;
    }

    return tmp;
}
NETLIB_EXPORTED(ds485_server_init);

void ds485_server_cleanup(struct dS485_server *server)
{
    pthread_rwlock_destroy(&server->clientListRWLock);
    free(server);
}
NETLIB_EXPORTED(ds485_server_cleanup);
